#include "stdafx.h"
#include "XDragUtil.h"
#include <shlobj.h>


CDragUtil::CDragUtil(FORMATETC* _pFormats,STGMEDIUM* _pStgMediums, int _formatNumber){
	refcount=1;
	formatNumber=_formatNumber;
	pFormats=new FORMATETC[formatNumber];    
	pStgMediums=new STGMEDIUM[formatNumber];     
	for(int i=0;i<formatNumber;++i){        
		pFormats[i]=_pFormats[i];        
		pStgMediums[i]=_pStgMediums[i];
	}
}
CDragUtil::~CDragUtil (){
	if(pFormats){
		delete []pFormats;
		pFormats=NULL;
	}
	if(pStgMediums){
		delete []pStgMediums;
		pStgMediums=NULL;
	}
}
STDMETHODIMP CDragUtil::QueryInterface(REFIID classid, void** intf){
	if(classid == IID_IUnknown){
		*intf = (IUnknown*)(IDataObject*)this;
		AddRef();
		return S_OK;
	}
	else if (classid == IID_IDataObject){
		*intf = (IDataObject*)this;
		AddRef();
		return S_OK;
	}
	else if (classid == IID_IDropSource){
		*intf = (IDropSource*)this;
		AddRef();
		return S_OK;
	}
	else return E_NOINTERFACE;
}
ULONG STDMETHODCALLTYPE CDragUtil::AddRef(){
	InterlockedIncrement(&refcount);
	return refcount;
}
ULONG STDMETHODCALLTYPE CDragUtil::Release(){
	InterlockedDecrement(&refcount);
	return refcount;
}
STDMETHODIMP CDragUtil::GetData(/* [unique][in] */ FORMATETC *pformatetcIn,/* [out] */ STGMEDIUM *pMedium){
	int index=LookupFormatEtc(pformatetcIn);  
	if(index<0)	return S_FALSE; 
	pMedium->tymed=pFormats[index].tymed;
	pMedium->pUnkForRelease=0;
	switch(pFormats[index].tymed){    
		case TYMED_HGLOBAL:
			pMedium->hGlobal=pStgMediums[index].hGlobal?DupGlobalMem(pStgMediums[index].hGlobal):NULL;        
			break;
		default:
			return S_FALSE;    
	}
	return S_OK;
}
STDMETHODIMP CDragUtil::GetDataHere(/* [unique][in] */ FORMATETC *pformatetc,/* [out][in] */ STGMEDIUM *pmedium){
	return DATA_E_FORMATETC;
}
STDMETHODIMP CDragUtil::QueryGetData( /* [unique][in] */ FORMATETC *pformatetc){
	return (LookupFormatEtc(pformatetc)==-1)?S_FALSE:S_OK;
}
STDMETHODIMP CDragUtil::GetCanonicalFormatEtc(/* [unique][in] */ FORMATETC *pformatectIn,/* [out] */ FORMATETC *pformatetcOut){
	pformatetcOut->ptd=NULL;    
	return E_NOTIMPL;
}
STDMETHODIMP CDragUtil::SetData(/* [unique][in] */ FORMATETC *pformatetc,/* [unique][in] */ STGMEDIUM *pmedium,/* [in] */ BOOL fRelease){
	return E_NOTIMPL;
}
STDMETHODIMP CDragUtil::EnumFormatEtc(/* [in] */ DWORD dwDirection,/* [out] */ IEnumFORMATETC **ppEnumFormatEtc){
	if(dwDirection==DATADIR_GET) return SHCreateStdEnumFmtEtc(formatNumber,pFormats,ppEnumFormatEtc);
	else return E_NOTIMPL;
}
STDMETHODIMP CDragUtil::DAdvise(/* [in] */ FORMATETC *pformatetc,/* [in] */ DWORD advf,/* [unique][in] */ IAdviseSink *pAdvSink,/* [out] */ DWORD *pdwConnection){
	return OLE_E_ADVISENOTSUPPORTED;
}
STDMETHODIMP CDragUtil::DUnadvise(/* [in] */ DWORD dwConnection){
	return OLE_E_ADVISENOTSUPPORTED;
}
STDMETHODIMP CDragUtil::EnumDAdvise(/* [out] */ IEnumSTATDATA **ppenumAdvise){
	return OLE_E_ADVISENOTSUPPORTED;
}
STDMETHODIMP CDragUtil::QueryContinueDrag(/* [in] */ BOOL fEscapePressed,/* [in] */ DWORD grfKeyState){
	HWND hWnd=GetCapture();
	if(!hWnd) return DRAGDROP_S_CANCEL;
	if(fEscapePressed == TRUE) return DRAGDROP_S_CANCEL;           
	if((grfKeyState & MK_LBUTTON) == 0) return DRAGDROP_S_DROP;
	return S_OK;
}
STDMETHODIMP CDragUtil::GiveFeedback(/* [in] */ DWORD dwEffect){
	return DRAGDROP_S_USEDEFAULTCURSORS;
}
int CDragUtil::LookupFormatEtc(FORMATETC *pFormatEtc){
	for(int i = 0; i < formatNumber; i++)    {
		if((pFormats[i].tymed    &  pFormatEtc->tymed)   &&
			pFormats[i].cfFormat == pFormatEtc->cfFormat &&
			pFormats[i].dwAspect == pFormatEtc->dwAspect)        {         
				return i;        
		}    
	}
	return -1;
}
HGLOBAL DupGlobalMem (HGLOBAL hMem){
	SIZE_T   len    = GlobalSize (hMem);
	PVOID   source = GlobalLock (hMem);
	PVOID   dest   = GlobalAlloc (GMEM_FIXED, len);
	memcpy (dest, source, len);
	GlobalUnlock (hMem);
	return dest;
}
DWORD CDragUtil::DoDragDrop(CLIPFORMAT cf,HGLOBAL hGLOBAL,DWORD dwOKEffects){
	FORMATETC fmtetc = {cf, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};
	STGMEDIUM stgmed = {TYMED_HGLOBAL, {0}, 0};
	stgmed.hGlobal=hGLOBAL;
	CDragUtil* pDU=new CDragUtil(&fmtetc, &stgmed, 1);
	DWORD ret=0;
	::DoDragDrop(pDU,pDU,dwOKEffects,&ret);
	delete pDU;
	return ret;
}
DWORD CDragUtil::DoDragDropText(const wchar_t* text,DWORD dwOKEffects){
	HGLOBAL hGLOBAL=Text2HGLOBAL(text);
	DWORD ret=DoDragDrop(CF_UNICODETEXT,hGLOBAL,dwOKEffects);
	GlobalFree(hGLOBAL);
	return ret;
}
DWORD CDragUtil::DoDragDropFiles(const wchar_t** pFilePaths,int fileNumber,DWORD dwOKEffects){
	HGLOBAL hGLOBAL=FilePaths2HGLOBAL(pFilePaths,fileNumber);
	DWORD ret=DoDragDrop(CF_HDROP,hGLOBAL,dwOKEffects);
	GlobalFree(hGLOBAL);
	return ret;
}
HGLOBAL CDragUtil::Text2HGLOBAL(const wchar_t* text){
	size_t l=(wcslen(text)+1)*sizeof(wchar_t);
	HGLOBAL hGLOBAL=GlobalAlloc (GMEM_FIXED, l);
	memcpy(hGLOBAL,text,l);
	return hGLOBAL;
}
HGLOBAL CDragUtil::FilePaths2HGLOBAL(const wchar_t** pFilePaths,int fileNumber){
	int* pSizes=(int*)GlobalAlloc(GMEM_FIXED,sizeof(int)*fileNumber);
	size_t l=sizeof(DROPFILES)+2;
	for(int i=0;i<fileNumber;++i){
		pSizes[i]=wcslen(pFilePaths[i]);
		l+=(pSizes[i]+1)*sizeof(wchar_t);
	}
	HGLOBAL hGLOBAL=GlobalAlloc (GMEM_FIXED, l);
	memset(hGLOBAL,0,l);
	DROPFILES* pDROPFILES=(DROPFILES*)hGLOBAL;
	pDROPFILES->fWide=TRUE;
	pDROPFILES->pFiles=sizeof(DROPFILES);
	pDROPFILES->fNC=TRUE;
	GetCursorPos(&(pDROPFILES->pt));
	BYTE* pTmp=(BYTE*)(++pDROPFILES);
	for(int j=0;j<fileNumber;++j){
		size_t sl=(pSizes[j]+1)*sizeof(wchar_t);
		memcpy(pTmp,pFilePaths[j],sl);
		pTmp+=sl;
	}
	GlobalFree((HGLOBAL)pSizes);
	return hGLOBAL;
}
